// Copyright (c) 2012, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

library compiler;

import 'dart:coreimpl';
import 'package:html5lib/dom.dart';
import 'package:html5lib/parser.dart';

import 'analyzer.dart';
import 'code_printer.dart';
import 'codegen.dart' as codegen;
import 'emitters.dart';
import 'messages.dart';
import 'file_system.dart';
import 'file_system/path.dart';
import 'files.dart';
import 'info.dart';
import 'options.dart';
import 'utils.dart';


/**
 * Parses an HTML file [contents] and returns a DOM-like tree.
 * Note that [contents] will be a [String] if coming from a browser-based
 * [FileSystem], or it will be a [List<int>] if running on the command line.
 */
Document parseHtml(contents, Path sourcePath) {
  var parser = new HtmlParser(contents, generateSpans: true);
  var document = parser.parse();

  // Note: errors aren't fatal in HTML (unless strict mode is on).
  // So just print them as warnings.
  for (var e in parser.errors) {
    messages.warning(e.message, e.span, file: sourcePath);
  }
  return document;
}

/** Compiles an application written with Dart web components. */
class Compiler {
  final FileSystem filesystem;
  final CompilerOptions options;
  final List<SourceFile> files = <SourceFile>[];
  final List<OutputFile> output = <OutputFile>[];

  PathInfo _pathInfo;

  /** Information about source [files] given their href. */
  final Map<Path, FileInfo> info = new SplayTreeMap<Path, FileInfo>();

  /** Used by template tool to open a file. */
  Compiler(this.filesystem, this.options);

  /** Compile the application starting from the given [mainFile]. */
  Future run(String mainFile, String outputDir, [String baseDir]) {
    var mainPath = new Path(mainFile);
    var basePath = baseDir != null ? new Path(baseDir) : mainPath.directoryPath;
    var outDir = new Path(outputDir);
    _pathInfo = new PathInfo(basePath, outDir);
    return _parseAndDiscover(mainPath).transform((_) {
      _analyze();
      _emit();
      return null;
    });
  }

  /**
   * Asynchronously parse [inputFile] and transitively discover web components
   * to load and parse. Returns a future that completes when all files are
   * processed.
   */
  Future _parseAndDiscover(Path inputFile) {
    var tasks = new FutureGroup();
    bool isEntry = true;

    processHtmlFile(SourceFile file) {
      files.add(file);

      var fileInfo = _time('Analyzed definitions', file.path,
          () => analyzeDefinitions(file, isEntryPoint: isEntry));
      isEntry = false;
      info[file.path] = fileInfo;

      // Load component files referenced by [file].
      for (var href in fileInfo.componentLinks) {
        tasks.add(_parseHtmlFile(href).transform(processHtmlFile));
      }

      // Load .dart files being referenced in the page.
      var src = fileInfo.externalFile;
      if (src != null) tasks.add(_parseDartFile(src).transform(_addDartFile));

      // Load .dart files being referenced in components.
      for (var component in fileInfo.declaredComponents) {
        var src = component.externalFile;
        if (src != null) tasks.add(_parseDartFile(src).transform(_addDartFile));
      }
    }

    tasks.add(_parseHtmlFile(inputFile).transform(processHtmlFile));
    return tasks.future;
  }

  /** Asynchronously parse [path] as an .html file. */
  Future<SourceFile> _parseHtmlFile(Path path) {
    return filesystem.readTextOrBytes(path).transform((source) {
      var file = new SourceFile(path);
      file.document = _time('Parsed', path, () => parseHtml(source, path));
      return file;
    });
  }

  /** Parse [filename] and treat it as a .dart file. */
  Future<SourceFile> _parseDartFile(Path path) {
    return filesystem.readText(path).transform(
        (source) => new SourceFile(path, isDart: true)..code = source);
  }

  void _addDartFile(SourceFile dartFile) {
    var fileInfo = new FileInfo(dartFile.path);
    info[dartFile.path] = fileInfo;
    fileInfo.inlinedCode = dartFile.code;
    files.add(dartFile);
  }

  /** Run the analyzer on every input html file. */
  void _analyze() {
    for (var file in files) {
      if (file.isDart) continue;
      _time('Analyzed contents', file.path, () => analyzeFile(file, info));
    }
  }

  /** Emit the generated code corresponding to each input file. */
  void _emit() {
    for (var file in files) {
      _time('Codegen', file.path, () {
        if (!file.isDart) {
          _removeScriptTags(file.document);
          _emitComponents(file);

          var fileInfo = info[file.path];
          if (fileInfo.isEntryPoint && fileInfo.codeAttached) {
            _emitMainDart(file);
            _emitMainHtml(file);
          } else {
            _emitComponentDart(file);
            _emitComponentHtml(file);
          }
        }
      });
    }
  }

  static const String DARTJS_LOADER =
    "http://dart.googlecode.com/svn/branches/bleeding_edge/dart/client/dart.js";

  /** Emit the main .dart file. */
  void _emitMainDart(SourceFile file) {
    var fileInfo = info[file.path];
    var contents = new MainPageEmitter(fileInfo).run(file.document);
    output.add(new OutputFile(_pathInfo.outputLibraryPath(fileInfo), contents));
  }

  /** Generate an html file with the (trimmed down) main html page. */
  void _emitMainHtml(SourceFile file) {
    var fileInfo = info[file.path];

    // Clear the body, we moved all of it
    var document = file.document;
    document.body.nodes.clear();

    output.add(new OutputFile(
        _pathInfo.fileInOutputDir('_${file.path.filename}_bootstrap.dart'),
        codegen.bootstrapCode(_pathInfo.relativePathFromOutputDir(fileInfo))));

    document.body.nodes.add(parseFragment(
      '<script type="text/javascript" src="$DARTJS_LOADER"></script>\n'
      '<script type="application/dart"'
      ' src="_${file.path.filename}_bootstrap.dart">'
      '</script>'
    ));

    for (var link in document.head.queryAll('link')) {
      if (link.attributes["rel"] == "components") {
        link.remove();
      }
    }

    _addAutoGeneratedComment(file);
    output.add(new OutputFile(
        _pathInfo.fileInOutputDir('_${file.path.filename}.html'),
        document.outerHTML));
  }

  /** Emits the Dart code for all components in the [file]. */
  void _emitComponents(SourceFile file) {
    var fileInfo = info[file.path];
    for (var component in fileInfo.declaredComponents) {
      var code = new WebComponentEmitter(fileInfo).run(component);
      output.add(new OutputFile(_pathInfo.outputLibraryPath(component), code));
    }
  }

  /** Emit the wrapper .dart file for a component page. */
  void _emitComponentDart(SourceFile file) {
    var fileInfo = info[file.path];
    // Reexport all components declared in the input file.
    var exports = fileInfo.declaredComponents.map(
        (c) => PathInfo.relativePath(fileInfo, c));
    output.add(new OutputFile(_pathInfo.outputLibraryPath(fileInfo),
        codegen.wrapComponentsLibrary(fileInfo, exports)));
  }

  /** Generate an html file declaring a web component. */
  void _emitComponentHtml(SourceFile file) {
    _addAutoGeneratedComment(file);
    output.add(new OutputFile(_pathInfo.outputPath(file.path, '.html'),
          file.document.outerHTML));
  }


  void _removeScriptTags(Document doc) {
    for (var tag in doc.queryAll('script')) {
      if (tag.attributes['type'] == 'application/dart') {
        tag.remove();
      }
    }
  }

  _time(String logMessage, Path path, callback(), {bool printTime: false}) {
    var message = new StringBuffer();
    message.add(logMessage);
    for (int i = (60 - logMessage.length - path.filename.length); i > 0 ; i--) {
      message.add(' ');
    }
    message.add(path.filename);
    return time(message.toString(), callback,
        printTime: options.verbose || printTime);
  }
}

void _addAutoGeneratedComment(SourceFile file) {
  var document = file.document;

  // Insert the "auto-generated" comment after the doctype, otherwise IE will go
  // into quirks mode.
  int commentIndex = 0;
  DocumentType doctype = find(document.nodes, (n) => n is DocumentType);
  if (doctype != null) {
    commentIndex = document.nodes.indexOf(doctype) + 1;
    // TODO(jmesserly): the html5lib parser emits a warning for missing doctype,
    // but it allows you to put it after comments. Presumably they do this
    // because some comments won't force IE into quirks mode (sigh). See this
    // link for more info:
    //     http://bugzilla.validator.nu/show_bug.cgi?id=836
    // For simplicity we're emitting the warning always, like validator.nu does.
    if (doctype.tagName != 'html' || commentIndex != 1) {
      messages.warning('file should start with <!DOCTYPE html> '
          'to avoid the possibility of it being parsed in quirks mode in IE. '
          'See http://www.w3.org/TR/html5-diff/#doctype',
          doctype.span, file: file.path);
    }
  }
  document.nodes.insertAt(commentIndex, parseFragment(
      '\n<!-- This file was auto-generated from template ${file.path}. -->\n'));
}
